unit Voxel_Tools;

interface

Uses Voxel;

// returns the number of redundant voxels removed
function RemoveRedundantVoxels(voxel: TVoxelSection): integer;

const FACING_LEFT = 2;
      FACING_RIGHT = FACING_LEFT * 2;
      FACING_TOP = FACING_RIGHT * 2;
      FACING_BOTTOM = FACING_TOP * 2;
      FACING_FRONT = FACING_BOTTOM * 2;
      FACING_BACK = FACING_FRONT * 2;
      FACING_MASK = FACING_LEFT or FACING_RIGHT or FACING_TOP or FACING_BOTTOM or FACING_FRONT or FACING_BACK;

// fills the Flags attribute of each voxel with facings information
// facings can be decoded by applying the above masks
// before running, always ensure that redundant voxels have been removed,
// else results will be screwed
procedure CalcFacings(voxel: TVoxelSection);

type FacingReport = array[0..255,0..255] of integer;

function GenerateFacingReport(voxel: TVoxelSection): FacingReport;

function FacingUsed(facing: integer; var report: FacingReport): boolean;

procedure PrintFacingReport(var report: FacingReport; var outfile: Text);
procedure PrintFacing(facing: integer; var report: FacingReport; var outfile: Text);

function FacingAsStr(facing: integer): string;

implementation

uses SysUtils;

function BitSet(value, mask: integer): boolean;
begin
     Result := (value and mask) = mask;
end;

const isInside = 0;
      isOutside = 1;

procedure FindInside(voxel: TVoxelSection);
{ This function starts by assuming every cell is inside.
  It then does a flood fill from every face of the cube,
  flood-filling all empty cells to outside.
}
var maxx, maxy, maxz: integer;
procedure ClearFlags;
var v: TVoxelUnpacked;
    x, y, z: integer;
begin
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do begin
                 voxel.GetVoxel(x,y,z,v);
                 v.Flags := isInside;
                 voxel.SetVoxel(x,y,z,v);
             end;
end;
var OutsideCount: integer;
procedure SetOutside(x,y,z: integer); // recursive- like a flood-fill
var v: TVoxelUnpacked;
begin
     // bounds check!
     if (x < 0) or (x > maxx) then Exit;
     if (y < 0) or (y > maxy) then Exit;
     if (z < 0) or (z > maxz) then Exit;
     // get the voxel
     voxel.GetVoxel(x,y,z,v);
     // already scanned?
     if v.Flags = isOutside then Exit;
     // is solid? stops a flood-fill
     if v.Used then Exit;
     // set to outside
     v.Flags := isOutside;
     Inc(OutsideCount); // record count
     voxel.SetVoxel(x,y,z,v);
     // and recurse on each axis
     SetOutside(x - 1, y, z);
     SetOutside(x + 1, y, z);
     SetOutside(x, y - 1, z);
     SetOutside(x, y + 1, z);
     SetOutside(x, y, z - 1);
     SetOutside(x, y, z + 1);
end;
var x, y, z: integer;
begin
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     ClearFlags;
     OutsideCount := 0;
     // fill each X side
     for y := 0 to maxy do
         for z := 0 to maxz do begin
             SetOutside(0,y,z); // will not fill used voxels anyway
             SetOutside(maxx,y,z);
         end;
     // fill each Y side
     for x := 0 to maxx do
         for z := 0 to maxz do begin
             SetOutside(x,0,z);
             SetOutside(x,maxy,z);
         end;
     // fill each Z side
     for x := 0 to maxx do
         for y := 0 to maxy do begin
             SetOutside(x,y,0);
             SetOutside(x,y,maxz);
         end;
//   x := (maxx+1)*(maxy+1)*(maxz+1);
//   ShowMessage(IntToStr(OutsideCount) + ' cells are outside' + Chr(10) +
//   IntToStr(x) + ' cells in total');
end;

function RemoveRedundantVoxels(voxel: TVoxelSection): integer;
var maxx, maxy, maxz: integer;
function IsRedundant(x,y,z: integer): boolean;
function IsOnOutside(x,y,z: integer): boolean;
var v: TVoxelUnpacked;
begin
     Result := True; // assume it is
     // bounds check
     if (x < 0) or (x > maxx) then Exit;
     if (y < 0) or (y > maxy) then Exit;
     if (z < 0) or (z > maxz) then Exit;
     // get the voxel
     voxel.GetVoxel(x,y,z,v);
     // is outside?
     Result := (v.Flags = isOutside);
end;
begin
     Result := false; // assume it isn't
     if IsOnOutside(x,y,z) then Exit;
     // ok, examine neighbours
     if IsOnOutside(x - 1,y,z) then Exit;
     if IsOnOutside(x + 1,y,z) then Exit;
     if IsOnOutside(x,y - 1,z) then Exit;
     if IsOnOutside(x,y + 1,z) then Exit;
     if IsOnOutside(x,y,z - 1) then Exit;
     if IsOnOutside(x,y,z + 1) then Exit;
     // if here, must be surrounded by insides
     Result := true;
end;
var v: TVoxelUnpacked;
    x, y, z: integer;
begin
     Result := 0; // none removed yet
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     // find the inside voxels
     FindInside(voxel);
     // examine each used voxel.  If it is beside all used inside voxels, then delete it
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do
                 if IsRedundant(x,y,z) then begin
                    voxel.GetVoxel(x,y,z,v);
                    if v.Used then begin // ignore already empty cells
                       v.Used := false; // delete it!
                       voxel.SetVoxel(x,y,z,v);
                       Inc(Result); // one more removed
                    end;
                 end;
     // done!
end;

procedure CalcFacings(voxel: TVoxelSection);
var maxx, maxy, maxz,
    x, y, z: integer;
procedure CalcFacing;
function Empty(x,y,z: integer): boolean;
var v: TVoxelUnpacked;
begin
     // check bounds
     Result := True; // outside the voxel is all empty
     if (x < 0) or (x > maxx) then Exit;
     if (y < 0) or (y > maxy) then Exit;
     if (z < 0) or (z > maxz) then Exit;
     // ok, do the real check
     voxel.GetVoxel(x,y,z,v);
     if v.Used then
        Result := False
     else
         Result := BitSet(v.Flags,isOutside);
end;
var v: TVoxelUnpacked;
    Flags: byte;
begin
     // get the voxel in question
     voxel.GetVoxel(x,y,z,v);
     // skip empty ones
     if not v.Used then Exit;
     Flags := v.Flags;
     // facings
     if Empty(x-1,y,z) then Flags := (Flags or FACING_FRONT);
     if Empty(x+1,y,z) then Flags := (Flags or FACING_BACK);
     if Empty(x,y-1,z) then Flags := (Flags or FACING_LEFT);
     if Empty(x,y+1,z) then Flags := (Flags or FACING_RIGHT);
     if Empty(x,y,z-1) then Flags := (Flags or FACING_BOTTOM);
     if Empty(x,y,z+1) then Flags := (Flags or FACING_TOP);
     // cancel out double-sided facings
     // TODO: if x is less than half-way, prefer front, else back etc
     if BitSet(Flags,FACING_FRONT or FACING_BACK) then Dec(Flags,FACING_BACK);
     if BitSet(Flags,FACING_LEFT or FACING_RIGHT) then Dec(Flags,FACING_RIGHT);
     if BitSet(Flags,FACING_TOP or FACING_BOTTOM) then Dec(Flags,FACING_BOTTOM);
     // save the result
     v.Flags := Flags;
     voxel.SetVoxel(x,y,z,v);
end;
begin
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     // examine each used voxel.
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do
                 CalcFacing;
end;

function GenerateFacingReport(voxel: TVoxelSection): FacingReport;
procedure Clear;
var facing, normal: integer;
begin
     for facing := 0 to 255 do
         for normal := 0 to 255 do
             Result[facing,normal] := 0;
end;
var x, y, z,
    maxx, maxy, maxz: integer;
    v: TVoxelUnpacked;
begin
     Clear;
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     // examine each used voxel.
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do begin
                 voxel.GetVoxel(x,y,z,v);
                 if v.Used then
                    Inc(Result[v.Flags and FACING_MASK,v.Normal]);
             end;
end;

function NormalsCount(facing: integer; var report: FacingReport): integer;
var normal: integer;
begin
     Result := 0;
     for normal := 0 to 255 do
         Inc(Result,report[facing,normal]);
end;

function FacingUsed(facing: integer; var report: FacingReport): boolean;
var normal, count: integer;
begin
     count := 0;
     for normal := 0 to 255 do
         Inc(count,report[facing,normal]);
     Result := (count > 0);
end;

procedure PrintFacing(facing: integer; var report: FacingReport; var outfile: Text);
var normal, sum, percent, astrix: integer;
begin
     sum := NormalsCount(facing,report);
     WriteLn(outfile,'facing ',facing shr 1,' ',FacingAsStr(facing),' (');
     for normal := 0 to 255 do
         if report[facing,normal] > 0 then begin // used?
            // data
            Write(outfile,#9,#9,normal:3,' = ',report[facing,normal]:3,'  ');
            // percent bar
            percent := Round((report[facing,normal] / sum) * 50); // 1 astrix per percent
            for astrix := 1 to percent do
                Write(outfile,'*');
            WriteLn(outfile);
         end;
     WriteLn(outfile,#9,')');
end;

procedure PrintFacingReport(var report: FacingReport; var outfile: Text);
var facing: integer;
begin
     for facing := 0 to 255 do begin
         if FacingUsed(facing,report) then
            PrintFacing(facing,report,outfile);
     end;
end;

function FacingAsStr(facing: integer): string;
begin
     Result := '';
     if BitSet(facing,FACING_LEFT) then Result := Concat(Result,' LEFT');
     if BitSet(facing,FACING_RIGHT) then Result := Concat(Result,' RIGHT');
     if BitSet(facing,FACING_TOP) then Result := Concat(Result,' TOP');
     if BitSet(facing,FACING_BOTTOM) then Result := Concat(Result,' BOTTOM');
     if BitSet(facing,FACING_FRONT) then Result := Concat(Result,' FRONT');
     if BitSet(facing,FACING_BACK) then Result := Concat(Result,' BACK');
     Result := Trim(Result);
     if Result = '' then Result := 'CONFUSED';
end;

end.
